
from tminterface.structs import BFEvaluationInfo, BFEvaluationResponse
from tminterface.constants import DEFAULT_SERVER_SIZE
import signal
import time


class Client(object):
    def __init__(self):
        pass

    def on_registered(self, iface):
        """
        A callback that the client has registered to a TMInterface instance.

        Args:
            iface (TMInterface): the TMInterface object that has been registered
        """
        pass

    def on_deregistered(self, iface):
        """
        A callback that the client has been deregistered from a TMInterface instance.
        This can be emitted when the game closes, the client does not respond in the timeout window,
        or the user manually deregisters the client with the `deregister` command.

        Args:
            iface (TMInterface): the TMInterface object that has been deregistered
        """
        pass

    def on_shutdown(self, iface):
        """
        A callback that the TMInterface server is shutting down. This is emitted when the game is closed.

        Args:
            iface (TMInterface): the TMInterface object that has been closed
        """
        pass

    def on_run_step(self, iface, _time: int):
        """
        Called on each "run" step (physics tick). This method will be called only in normal races and not
        when validating a replay.

        Args:
            iface (TMInterface): the TMInterface object
        """
        pass

    def on_simulation_begin(self, iface):
        """
        Called when a new simulation session is started (when validating a replay).

        Args:
            iface (TMInterface): the TMInterface object
        """
        pass

    def on_simulation_step(self, iface, _time: int):
        """
        Called on each simulation step (physics tick). This method will be called only when validating a replay.

        Args:
            iface (TMInterface): the TMInterface object
        """
        pass
    
    def on_simulation_end(self, iface, result: int):
        """
        Called when a new simulation session is ended (when validating a replay).

        Args:
            iface (TMInterface): the TMInterface object
        """
        pass

    def on_checkpoint_count_changed(self, iface, current: int, target: int):
        """
        Called when the current checkpoint count changed (a new checkpoint has been passed by the vehicle).
        The `current` and `target` parameters account for the total amount of checkpoints to be collected,
        taking lap count into consideration.

        Args:
            iface (TMInterface): the TMInterface object
            current (int): the current amount of checkpoints passed
            target (int): the total amount of checkpoints on the map (including finish)
        """
        pass

    def on_laps_count_changed(self, iface, current: int):
        """
        Called when the current lap count changed (a new lap has been passed).

        Args:
            iface (TMInterface): the TMInterface object
            current (int): the current amount of laps passed
        """
        pass

    def on_custom_command(self, iface, time_from: int, time_to: int, command: str, args: list):
        """
        Called when a custom command has been executed by the user.

        Args:
            iface (TMInterface): the TMInterface object
            time_from (int): if provided by the user, the starting time of the command, otherwise -1
            time_to (int): if provided by the user, the ending time of the command, otherwise -1
            command (str): the command name being executed
            args (list): the argument list provided by the user
        """
        pass

    def on_bruteforce_evaluate(self, iface, info: BFEvaluationInfo) -> BFEvaluationResponse:
        """
        Called on each bruteforce physics step iteration. This method will only be called when
        the bruteforce script is enabled in TMInterface. Used for implementing custom evaluation
        strategies. For greater control over the simulation, use the Client.on_simulation_step method instead.

        Args:
            iface (TMInterface): the TMInterface object
            info (BFEvaluationInfo): the info about the current bruteforce settings and race time

        Returns:
            None if the bruteforce script should continue its builtin evaluation or a BFEvaluationResponse
            that signifies what the script should do.
        """
        return None

    def on_client_exception(self, iface, exception: Exception):
        """
        Called when a client exception is thrown. This can happen if opening the shared file fails, or reading from
        it fails.

        Args:
            iface (TMInterface): the TMInterface object
            exception (Exception): the exception being thrown
        """
        print(f'[Client] Exception reported: {exception}')


def run_client(client: Client, server_name: str = 'TMInterface0', buffer_size=DEFAULT_SERVER_SIZE):
    """
    Connects to a server with the specified server name and registers the client instance.
    The function closes the connection on SIGBREAK and SIGINT signals and will block
    until the client is deregistered in any way. You can set the buffer size yourself to use for
    the connection, by specifying the buffer_size parameter. Using a custom size requires
    launching TMInterface with the /serversize command line parameter: TMInterface.exe /serversize=size.

    Args:
        client (Client): the client instance to register
        server_name (str): the server name to connect to, TMInterface0 by default
        buffer_size (int): the buffer size to use, the default size is defined by tminterface.constants.DEFAULT_SERVER_SIZE
    """
    from .interface import TMInterface

    iface = TMInterface(server_name, buffer_size)

    def handler(signum, frame):
        iface.close()

    signal.signal(signal.SIGBREAK, handler)
    signal.signal(signal.SIGINT, handler)

    iface.register(client)

    while iface.running:
        time.sleep(0)




import tminterface.util as util
from tminterface.constants import ANALOG_ACCELERATE_NAME, ANALOG_STEER_NAME, BINARY_ACCELERATE_NAME, BINARY_BRAKE_NAME, BINARY_HORN_NAME, BINARY_LEFT_NAME, BINARY_RACE_FINISH_NAME, BINARY_RACE_START_NAME, BINARY_RESPAWN_NAME, BINARY_RIGHT_NAME
from typing import Union
from math import ceil


class Event(object):
    """
    The Event class represents a game event (or input) with its respective time.

    Such event is stored in 8 bytes internally by the game. The first 4 bytes is the time
    of the event. This time is a stored time, which means it is offset by 100000ms.

    The last 4 bytes contain the event data. This data contains the actual event type
    (e.g. whether it was acceleration, braking, steering etc.) and the value of the event.

    The event type is 1 byte long and it signifes an index into an array of available event types.
    This array is variable based on the information contained within the replay file. As such,
    it is required to get index of the desired event type dynamically. You can easily get/set this property
    through the name_index accessors.

    The event value depends on the event type and is 3 bytes long. If the event type is binary
    (e.g. accelerate, brake, steer left), the value can be either 0 or 1. For managing this value type,
    use the binary_value getter/setter.

    If the event type is analog, the value is stored in a custom format. You can convert between this
    format and the format TMInterface uses by using util.data_to_analog_value and utils.analog_value_to_data.
    You can avoid using these functions simply by using the analog_value getter/setter.

    Args:
        time (int): the stored time of the event

    Attributes:
        time (int): the stored time of the event
        data (int): the final data that is written into game's memory
    """
    def __init__(self, time: int, data: int = 0):
        self.time = time
        self.data = data

    @property
    def name_index(self) -> int:
        return self.data >> 24

    @name_index.setter
    def name_index(self, index: int):
        self.data &= 0xFFFFFF
        self.data |= (index << 24)

    @property
    def binary_value(self) -> bool:
        return bool(self.data & 0xFFFFFF)

    @binary_value.setter
    def binary_value(self, value: bool):
        self.data = self.data & 0xFF000000 | int(value)

    @property
    def analog_value(self) -> int:
        return util.data_to_analog_value(self.data & 0xFFFFFF)

    @analog_value.setter
    def analog_value(self, value: int):
        self.data = self.data & 0xFF000000 | (util.analog_value_to_data(value) & 0xFFFFFF)


class EventBufferData(object):
    """
    The internal event buffer used to hold player inputs in run or simulation mode.

    While simulating a race, the game loads the inputs from a replay file
    into an internal buffer and begins to apply "events" (inputs) from this
    buffer. The buffer itself consists of 8 byte values, the first 4 bytes
    is used for the event time and the last 4 is used for event data.
    See the Event class for more information.

    The event time is so called a "stored" time. The stored time is
    defined as 100000 + race time. The stored time is saved in the
    replay file and is also used in the internal buffer itself.

    The buffer itself is stored in *decreasing* order. That means that the event
    at index 0 in the list is the last one simulated in the race. The start and end
    of the race is marked by special "_FakeIsRaceRunning" and "_FakeFinishLine" events.
    These events mark the start and finish of the race. Note that without the presence
    of "_FakeIsRaceRunning" event, the race will not start at all. This event has a
    constant stored time of 100000.

    Before the starting event, a "Respawn" event can be generated by the game, this
    event can also be saved in the replay file itself. The very first input that can be applied
    by the player happens at stored time of 100010.

    Arguments:
        events_duration (int): the duration of the events, equalling the finish time, mostly ignored and does not need to be set

    Attributes:
        events_duration (int): the duration of the events
        control_names (list): the list of supported event types by this buffer
        events (list): the list of events held by this buffer
    """
    def __init__(self, events_duration: int):
        self.events_duration = events_duration
        self.control_names = []
        self.events = []

    def copy(self):
        """
        Copies the event buffer with all its events.

        Returns:
            a deep copy of the original event buffer
        """
        cpy = EventBufferData(self.events_duration)
        cpy.control_names = self.control_names[:]
        cpy.events = self.events[:]
        return cpy

    def clear(self):
        """
        Removes all events in the current event buffer, leaving
        the race running event in the buffer.

        A race running event should always be present in the buffer, to
        make the game start the race.
        """
        self.events = []
        self.add(-10, '_FakeIsRaceRunning', True)
    
    def sort(self):
        """
        Sorts the event buffer by time in decreasing order.

        This is the order that events are stored internally by the game.
        Calling this is not needed, if you are calling set_event_buffer. 
        The server will always take care of properly sorting the events.
        """
        self.events = sorted(self.events, key=lambda ev: ev.time, reverse=True)

    def add(self, time: int, event_name: str, value: Union[int, bool]):
        """
        Adds an event to the event buffer.

        This is a wrapper function that provides easy API for adding new events.
        Depending on the event_name parameter, the method will interpret the value
        in different ways. If the event is an analog event, the value passed should
        be in the range of [-65536, 65536] where negative values represent left steering
        and postive, right steering.
        
        If the event is binary, the value should be False for disabling the input and 
        True for enabling it.

        The time parameter is zero based, where 0 is the first human input that can be injected.
        Internally, 0 translates to stored time 100010, which is the first simulation step
        after the countdown.

        Args:
            time (int): zero based timestamp when the input is injected
            event_name (str): the event name that specifies the input type
            value (Union[int, bool]): the value for the event, based on the event type
        """
        try:
            index = self.control_names.index(event_name)
        except ValueError:
            raise ValueError(f'Event name "{event_name}" does not exist in this event buffer')

        ev = Event(time + 100010)
        ev.name_index = index
        if event_name == ANALOG_ACCELERATE_NAME or event_name == ANALOG_STEER_NAME:
            ev.analog_value = value
        else:
            ev.binary_value = value

        self.events.append(ev)

    def find(self, **kwargs):
        """
        Finds matching events according to keyword arguments.

        Any unspecified parameter will be skipped in the search and will not be compared.
        You may use this method to filter events based on time, event type and value. 

        Find all analog steering events with value -65536::

            matching = event_buffer.find(event_name=structs.ANALOG_STEER_NAME, value=-65536)

        Find all events that happened at input time 0::

            matching = event_buffer.find(time=0)

        Find the finish line event::

            matching = event_buffer.find(event_name=structs.BINARY_RACE_FINISH_NAME, value=True)

        Calling this method without any keyword arguments will return all events in the buffer.

        Args:
            **kwargs: the keyword arguments
        
        Keyword Args:
            event_name (str): match events with this event type
            time (int): match events with this time (zero based)
            value (Union[int, bool]): match events with this value, bool if the event type is binary, int if analog,
                this parameter can only be filtered if event_name is provided

        Returns:
            list: the events that matched the query
        """
        index = -1
        if 'event_name' in kwargs:
            try:
                index = self.control_names.index(kwargs['event_name'])
            except ValueError:
                raise ValueError(f'Event name "{kwargs["event_name"]}" does not exist in this event buffer')

        has_value = 'value' in kwargs
        has_time = 'time' in kwargs

        matched = []
        for ev in reversed(self.events):
            if has_time and ev.time - 100010 != kwargs['time']:
                continue

            if index >= 0:
                if ev.name_index != index:
                    continue

                if has_value:
                    if kwargs['event_name'] == ANALOG_STEER_NAME or kwargs['event_name'] == ANALOG_ACCELERATE_NAME:
                        if ev.analog_value != kwargs['value']:
                            continue
                    else:
                        if ev.binary_value != kwargs['value']:
                            continue

            matched.append(ev)

        return matched

    def to_commands_str(self, all_events=False):
        """
        Converts event buffer events and constructs a string consisting
        of commands compatible with TMInterface's script syntax.

        By default, only events that happened in the race
        (signified by the race running and finish events) will be converted
        and appended to the final string. If you want to convert all commands
        that are available in the buffer, call this method with all_events set to True.

        Args:
            all_events (bool): whether to convert all commands available in the buffer
        
        Returns:
            str: the string containing commands compatible with TMInterface's script syntax
        """
        ACTION_MAPPINGS = {
            BINARY_ACCELERATE_NAME: 'up',
            BINARY_BRAKE_NAME: 'down',
            BINARY_LEFT_NAME: 'left',
            BINARY_RIGHT_NAME: 'right',
            BINARY_RESPAWN_NAME: 'enter',
            BINARY_HORN_NAME: 'horn'
        }

        sorted_events = sorted(self.events, key=lambda ev: ev.time)

        commands = ''
        try:
            start_events = self.find(event_name=BINARY_RACE_START_NAME)
            if start_events:
                start_time = start_events[0].time
            else:
                start_time = 100000
        except ValueError:
            start_time = 100000

        for ev in sorted_events:
            event_name = self.control_names[ev.name_index]
            if not all_events:
                if ev.time < start_time:
                    continue

                if event_name == BINARY_RACE_FINISH_NAME:
                    break

            if event_name in [BINARY_RESPAWN_NAME, BINARY_HORN_NAME] and not ev.binary_value:
                continue

            time = int(ceil((ev.time - start_time - 10) / 10) * 10)
            if event_name in ACTION_MAPPINGS.keys():
                if ev.binary_value:
                    commands += f'{time} press {ACTION_MAPPINGS[event_name]}\n'
                else:
                    commands += f'{time} rel {ACTION_MAPPINGS[event_name]}\n'
                
            elif event_name == ANALOG_ACCELERATE_NAME:
                commands += f'{time} gas {ev.analog_value}\n'
            elif event_name == ANALOG_STEER_NAME:
                commands += f'{time} steer {ev.analog_value}\n'
        
        return commands



from numpy import int32
import numpy as np
import math

EPSILON = 0.00001

def data_to_analog_value(data: int) -> int:
    """
    Converts an internal analog state value to a [-65536, 65536] range.

    The function supports values outside the normal range, that is
    you can convert values in the extended range as well.

    Args:
        data (int): the internal value, usually stored in an event buffer
    
    Returns:
        int: the converted value
    """
    val = int32(data)
    val <<= int32(8)
    val >>= int32(8)
    return -val

def analog_value_to_data(value: int) -> int:
    """
    Converts a value in [-65536, 65536] range to an internal analog state value.

    The function supports values outside the normal range, that is
    you can convert values in the extended range as well.

    Args:
        data (int): the value to convert
    
    Returns:
        int: the converted value
    """
    value = -value
    value <<= int32(8)
    value >>= int32(8)
    return value


def quat_to_ypw(quat: np.array) -> np.array:
    """
    Converts a quaternion to yaw, pitch and roll values.

    This function uses the internal implementation that the game
    uses to convert quaternions to yaw, pitch and roll.

    The function itself is constructed from reverse engineering
    the code of the game. Note that this particular implementation
    is not 100% compatible with the actual assembly code, however it
    should produce the same results in most cases.

    Args:
        quat (np.array): the quaternion to convert (x, y, z, w)

    Returns:
        np.array: an array containing 3 elements: yaw, pitch and roll
    """
    t0 = quat[2] * quat[1] + quat[3] * quat[0]

    if abs(t0 + 0.5) < EPSILON or t0 + 0.5 <= 0:
        yaw = math.atan2(quat[1], quat[0])
        return np.array([yaw * 2, -1.57079637, 0])

    if abs(t0 - 0.5) < EPSILON or t0 - 0.5 >= 0:
        yaw = math.atan2(quat[1], quat[0])
        return np.array([-yaw * 2, 1.57079637, 0])

    yaw = math.atan2(2.0 * (quat[2] * quat[0] - quat[3] * quat[1]), 1.0 - (quat[3] * quat[3] + quat[2] * quat[2]) * 2.0)
    roll = math.asin(2.0 * t0)
    pitch = math.atan2(2.0 * (quat[0] * quat[1] - quat[2] * quat[3]), 1.0 - 2.0 * (quat[1] * quat[1] + quat[3] * quat[3]))
    return np.array([yaw, pitch, roll])


def mat3_to_quat(mat: np.array) -> np.array:
    """
    Converts a rotation matrix to a quaternion.

    This function uses the internal implementation that the game
    uses to convert a rotation matrix to a quaternion.

    The function itself is constructed from reverse engineering
    the code of the game. This particular implementation should be
    100% compatible with the original method the game uses.
    Note however that this compatibility is not guaranteed.

    Args:
        mat (np.array): a 3x3 rotation matrix to convert

    Returns:
        np.array: the quaternion consisting of 4 elements (x, y, z, w)
    """
    trace = np.trace(mat)

    if trace > 0:
        trace += 1
        trace_squared = math.sqrt(trace)

        trace  = 0.5 / trace_squared
        return np.array([
            trace_squared / 2,
            trace * (mat[2, 1] - mat[1, 2]),
            trace * (mat[0, 2]  - mat[2, 0]),
            trace * (mat[1, 0] - mat[0, 1])
        ])

    index = 0
    if mat[1, 1] > mat[0, 0]:
        index = 1
    elif mat[2, 2] > mat[0, 0]:
        index = 2

    indexes = [1, 2, 0]
    var_1 = indexes[index]
    var_2 = indexes[var_1]

    trace = mat[index, index] - (mat[var_2, var_2] + mat[var_1, var_1]) + 1.0

    trace_squared = math.sqrt(trace)
    trace = 0.5 / trace_squared

    quat = np.zeros(4)
    quat[0] = trace * (mat[var_2, var_1] - mat[var_1, var_2])
    quat[index + 1] = trace_squared / 2
    quat[var_1 + 1] = trace * (mat[index, var_1] + mat[var_1, index])
    quat[var_2 + 1] = trace * (mat[index, var_2] + mat[var_2, index])
    return quat



TIMERS_SIZE = 212
CMD_BUFFER_CORE_SIZE = 264
DYNA_SIZE = 1424
SCENE_MOBIL_SIZE = 2168
SIMULATION_WHEELS_SIZE = 3056
PLUG_SOLID_SIZE = 68
PLAYER_INFO_SIZE = 952
INPUT_STATE_SIZE = 120

SIM_HAS_TIMERS = 0x1
SIM_HAS_DYNA = 0x2
SIM_HAS_SCENE_MOBIL = 0x4
SIM_HAS_SIMULATION_WHEELS = 0x8
SIM_HAS_PLUG_SOLID = 0x10
SIM_HAS_CMD_BUFFER_CORE = 0x20
SIM_HAS_INPUT_STATE = 0x40
SIM_HAS_PLAYER_INFO = 0x80

MODE_SIMULATION = 0
MODE_RUN = 1

BINARY_ACCELERATE_NAME = 'Accelerate'
BINARY_BRAKE_NAME = 'Brake'
BINARY_LEFT_NAME = 'Steer left'
BINARY_RIGHT_NAME = 'Steer right'
BINARY_RESPAWN_NAME = 'Respawn'
BINARY_HORN_NAME = 'Horn'
BINARY_RACE_START_NAME = '_FakeIsRaceRunning'
BINARY_RACE_FINISH_NAME = '_FakeFinishLine'
ANALOG_STEER_NAME = 'Steer (analog)'
ANALOG_ACCELERATE_NAME = 'Accelerate (analog)'

DEFAULT_SERVER_SIZE = 65535


import struct
from enum import IntEnum
from tminterface.constants import MODE_RUN, SIM_HAS_TIMERS, SIM_HAS_DYNA, SIM_HAS_PLAYER_INFO
from tminterface.eventbuffer import Event
from tminterface.util import mat3_to_quat, quat_to_ypw
import numpy as np


class CheckpointData(object):
    """
    The CheckpointData object represents checkpoint state within the game.

    The game keeps track of two arrays that contain checkpoint information.
    The first "state" array is an array of booleans (a boolean is 4 bytes long)
    and keeps track of which checkpoints were already passed. The length of the
    array represents the real count of the checkpoints on current the map (including finish).
    This does not mean that to finish the race the player has to pass through this exact count
    of checkpoints. A map with 3 laps and 5 checkpoints will still contain only 5 checkpoint states.

    The second "times" array is an array of structures with 2 fields: time and an unknown field.
    This array holds the "logical" number of checkpoints that have to be passed (including finish).
    This means the total number of checkpoint passes, including the existence of laps.

    Arguments:
        cp_states (list): the checkpoint states array
        cp_times (list): the checkpoint times array, each element is a two element tuple of (time, flags)
    """
    def __init__(self, cp_states: list, cp_times: list):
        self.cp_states = cp_states
        self.cp_times = cp_times


class SimStateData(object):
    """
    The SimStateData object represents a full save state of the simulation state,
    including checkpoint and input information.

    The simulation state consists of raw memory buffers representing various
    information about the race state. This includes the entirety of the vehicle
    state as well as the player info and other state variables such as current
    checkpoint count and input state.

    The memory regions themselves are monitored by TMInterface itself and are used
    for functionality like save states or fast rewind in the bruteforce script.
    TMInterface may use additional native game methods to restore the state based
    on information present in some of these memory regions. It is important to note
    that the buffers contain instance specific fields such as pointers and array sizes.
    These are masked out automatically by TMInterface when restoring the state
    (and when calling TMInterface.rewind_to_state).

    To query input state of the simulation state regardless of context,
    use input_* (input_accelerate, input_brake etc.) accessors.
    """
    def __init__(self):
        self.version = 0
        self.context_mode = MODE_RUN
        self.flags = 0
        self.timers = bytearray()
        self.dyna = bytearray()
        self.scene_mobil = bytearray()
        self.simulation_wheels = bytearray()
        self.plug_solid = bytearray()
        self.cmd_buffer_core = bytearray()
        self.player_info = bytearray()
        self.internal_input_state = bytearray()
        self.input_running_event = Event(0)
        self.input_finish_event = Event(0)
        self.input_accelerate_event = Event(0)
        self.input_brake_event  = Event(0)
        self.input_left_event = Event(0)
        self.input_right_event = Event(0)
        self.input_steer_event = Event(0)
        self.input_gas_event = Event(0)
        self.num_respawns = 0
        self.cp_data = None

    @property
    def time(self) -> int:
        if (self.flags & SIM_HAS_TIMERS) == 0:
            return 0

        return self.__get_int(self.timers, 4)

    @property
    def position(self) -> list:
        if (self.flags & SIM_HAS_DYNA) == 0:
            return [0, 0, 0]
        
        return self.__get_vec3(self.dyna, 500)

    @property
    def velocity(self) -> list:
        if (self.flags & SIM_HAS_DYNA) == 0:
            return [0, 0, 0]

        return self.__get_vec3(self.dyna, 512)

    # Available only in run context
    @property
    def display_speed(self) -> int:
        if (self.flags & SIM_HAS_PLAYER_INFO) == 0:
            return 0

        return self.__get_int(self.player_info, 832)

    @position.setter
    def position(self, pos: list) -> bool:
        if (self.flags & SIM_HAS_DYNA) == 0:
            return False

        self.__set_vec3(self.dyna, 500, pos)
        return True

    @velocity.setter
    def velocity(self, vel: list) -> bool:
        if (self.flags & SIM_HAS_DYNA) == 0:
            return False

        self.__set_vec3(self.dyna, 512, vel)
        return True

    @property
    def rotation_matrix(self) -> list:
        if (self.flags & SIM_HAS_DYNA) == 0:
            return [[0, 0, 0]] * 3

        m1 = self.__get_vec3(self.dyna, 464)
        m2 = self.__get_vec3(self.dyna, 476)
        m3 = self.__get_vec3(self.dyna, 488)

        return [m1, m2, m3]

    @rotation_matrix.setter
    def rotation_matrix(self, matrix: list) -> bool:
        if (self.flags & SIM_HAS_DYNA) == 0:
            return False

        self.__set_vec3(self.dyna, 464, matrix[0])
        self.__set_vec3(self.dyna, 476, matrix[1])
        self.__set_vec3(self.dyna, 488, matrix[2])

    @property
    def yaw_pitch_roll(self) -> np.array:
        if (self.flags & SIM_HAS_DYNA) == 0:
            return [0, 0, 0]

        mat = np.array(self.rotation_matrix)
        return list(quat_to_ypw(mat3_to_quat(mat)))

    @property
    def race_time(self) -> int:
        if (self.flags & SIM_HAS_PLAYER_INFO) == 0:
            return False

        return self.__get_int(self.player_info, 688)

    @property
    def rewind_time(self) -> int:
        return self.race_time + 10

    @property
    def input_accelerate(self) -> bool:
        return self.input_accelerate_event.data if self.context_mode == MODE_RUN else self.input_accelerate_event.binary_value

    @property
    def input_brake(self) -> bool:
        return self.input_brake_event.data if self.context_mode == MODE_RUN else self.input_brake_event.binary_value

    @property
    def input_left(self) -> bool:
        return self.input_left_event.data if self.context_mode == MODE_RUN else self.input_left_event.binary_value

    @property
    def input_right(self) -> bool:
        return self.input_right_event.data if self.context_mode == MODE_RUN else self.input_right_event.binary_value

    @property
    def input_steer(self) -> int:
        return self.input_steer_event.data if self.context_mode == MODE_RUN else self.input_steer_event.analog_value

    @property
    def input_gas(self) -> int:
        return self.input_gas_event.data if self.context_mode == MODE_RUN else self.input_gas_event.analog_value

    @staticmethod
    def __get_int(buffer: bytearray, offset: int) -> int:
        return int.from_bytes(buffer[offset:offset+4], byteorder='little')

    @staticmethod
    def __get_vec3(buffer: bytearray, offset: int) -> list:
        x = struct.unpack('f', buffer[offset:offset+4])[0]
        y = struct.unpack('f', buffer[offset+4:offset+8])[0]
        z = struct.unpack('f', buffer[offset+8:offset+12])[0]
        return [x, y, z]

    @staticmethod
    def __set_vec3(buffer: bytearray, offset: int, v: list):
        buffer[offset:offset+4] = list(struct.pack('f', v[0]))
        buffer[offset+4:offset+8] = list(struct.pack('f', v[1]))
        buffer[offset+8:offset+12] = list(struct.pack('f', v[2]))


class BFPhase(IntEnum):
    """
    The phase in which the bruteforce script is currently working.

    The initial phase is executed at the beginning of the process and after each improvement.
    It is used primarily for collecting data about the race e.g: the race time, position of the car,
    checkpoint times etc. No state modification happens at this state and it is recommended to use this phase
    to collect information about the current solution.

    The search phase is when TMInterface is searching for a new improvement. In this phase, the process
    changes inputs according to the user settings and evaluates the solution based on the current target.
    """
    INITIAL = 0
    SEARCH = 1


class BFTarget(IntEnum):
    """
    The bruteforce metric that is being currently optimized.
    """
    FINISH_TIME = 0
    CHECKPOINT_TIME = 1
    TRIGGER = 2


class BFEvaluationDecision(IntEnum):
    """
    The decision taken by the client in every bruteforce physics step.
    Returned in :meth:`Client.on_bruteforce_evaluate` in an BFEvaluationResponse instance.

    `CONTINUE`: run the default evaluation of the bruteforce script

    `DO_NOTHING`: do not run any evaluation that could result in accepting or rejecting the evaluated solution

    `ACCEPT`: accept the current solution as the new best one. Starts a new intial phase in the next physics step.

    `REJECT`: rejects the current solution and generates a new one for next evaluation

    `STOP`: stops the bruteforce script and lets the game simulate the race until the end
    """
    CONTINUE = 0
    DO_NOTHING = 1
    ACCEPT = 2
    REJECT = 3
    STOP = 4


class BFEvaluationInfo(object):
    """
    The bruteforce settings applied in the bruteforce process, including the current simulation race time.
    """
    def __init__(self) -> None:
        self.phase = BFPhase.INITIAL
        self.target = BFTarget.FINISH_TIME
        self.time = 0
        self.modified_inputs_num = -1
        self.inputs_min_time = -1
        self.inputs_max_time = -1
        self.max_steer_diff = -1
        self.max_time_diff = -1
        self.override_stop_time = -1
        self.search_forever = False
        self.inputs_extend_steer = False


class BFEvaluationResponse(object):
    """
    The response object sent by :meth:`Client.on_bruteforce_evaluate`.
    
    If `decision` is set to :class:`BFEvaluationDecision.REJECT`,
    you are allowed to change the inputs manually via the :meth:`TMInterface.set_event_buffer` method and
    set the `rewind_time` to `timestamp - 10` where `timestamp` is the first input that has been
    changed by your algorithm. Otherwise, TMInterface will automatically randomize the inputs according
    to the current settings itself.
    """
    def __init__(self) -> None:
        self.decision = BFEvaluationDecision.CONTINUE
        self.rewind_time = -1



import struct
import threading
import time
import mmap
from typing import Tuple

from tminterface.client import Client
from tminterface.structs import BFEvaluationResponse, BFEvaluationInfo, BFPhase, BFTarget, CheckpointData, SimStateData
from tminterface.eventbuffer import EventBufferData, Event
from tminterface.constants import *
from enum import IntEnum, auto


class MessageType(IntEnum):
    S_RESPONSE = auto()
    S_ON_REGISTERED = auto()
    S_SHUTDOWN = auto()
    S_ON_RUN_STEP = auto()
    S_ON_SIM_BEGIN = auto()
    S_ON_SIM_STEP = auto()
    S_ON_SIM_END = auto()
    S_ON_CHECKPOINT_COUNT_CHANGED = auto()
    S_ON_LAPS_COUNT_CHANGED = auto()
    S_ON_CUSTOM_COMMAND = auto()
    S_ON_BRUTEFORCE_EVALUATE = auto()
    C_REGISTER = auto()
    C_DEREGISTER = auto()
    C_PROCESSED_CALL = auto()
    C_SET_INPUT_STATES = auto()
    C_RESPAWN = auto()
    C_GIVE_UP = auto()
    C_HORN = auto()
    C_SIM_REWIND_TO_STATE = auto()
    C_SIM_GET_STATE = auto()
    C_SIM_GET_EVENT_BUFFER = auto()
    C_GET_CONTEXT_MODE = auto()
    C_SIM_SET_EVENT_BUFFER = auto()
    C_SIM_SET_TIME_LIMIT = auto()
    C_GET_CHECKPOINT_STATE = auto()
    C_SET_CHECKPOINT_STATE = auto()
    C_SET_GAME_SPEED = auto()
    C_EXECUTE_COMMAND = auto()
    C_SET_EXECUTE_COMMANDS = auto()
    C_SET_TIMEOUT = auto()
    C_REMOVE_STATE_VALIDATION = auto()
    C_PREVENT_SIMULATION_FINISH = auto()
    C_REGISTER_CUSTOM_COMMAND = auto()
    C_LOG = auto()
    ANY = auto()


RESPONSE_TOO_LONG = 1
CLIENT_ALREADY_REGISTERED = 2
NO_EVENT_BUFFER = 3
NO_PLAYER_INFO = 4
COMMAND_ALREADY_REGISTERED = 5

MAXINT32 = 2 ** 31 - 1


class Message(object):
    """
    The Message class represents a binary buffer that contains useful methods to construct
    a message to send to the server. A message additionally contains its type, whether it is
    a response to a server call, or a normal client call. It also contains an error code,
    if there was any failure writing the message.

    Args:
        _type (int): the message type
        error_code (int): the error code of the message, 0 if none

    Attributes:
        _type (int): the message type
        error_code (int): the error code of the message, 0 if none
        data (bytearray): the binary data
    """
    def __init__(self, _type: int, error_code=0):
        self._type = _type
        self.error_code = error_code
        self.data = bytearray()

    def write_event(self, event: Event):
        self.write_uint32(event.time)
        self.write_int32(event.data)

    def write_uint8(self, n):
        self.data.extend(struct.pack('B', n))

    def write_int16(self, n: int):
        self.data.extend(struct.pack('h', n))

    def write_uint16(self, n: int):
        self.data.extend(struct.pack('H', n))

    def write_int32(self, n: int):
        self.data.extend(struct.pack('i', n))

    def write_uint32(self, n: int):
        self.data.extend(struct.pack('I', n))

    def write_double(self, n: float):
        self.data.extend(struct.pack('d', n))

    def write_buffer(self, buffer: bytearray):
        self.data.extend(buffer)

    def write_zeros(self, n_bytes):
        self.data.extend(bytearray(n_bytes))

    def write_int(self, n, size):
        if size == 1:
            self.write_uint8(n)
        elif size == 2:
            if n < 0:
                self.write_int16(n)
            else:
                self.write_uint16(n)
        elif size == 4:
            if n == 0xffffffff:
                self.write_uint32(n)
            else:
                self.write_int32(n)

    def to_data(self) -> bytearray:
        return bytearray(struct.pack('i', self._type)) + bytearray(struct.pack('i', self.error_code)) + self.data

    def __len__(self):
        return 8 + len(self.data)


class ServerException(Exception):
    """
    An exception thrown when the server cannot perform requested operation.
    """
    pass


class TMInterface(object):
    """
    TMInterface is the main class to communicate with the TMInterface server.
    The communication is done through memory mapping and a simple synchronous
    message system between the server and the client. A TMInterface server
    can only serve only one client at a time, it is however possible to connect
    to many servers from the same script.

    The interface provides various functions to manipulate game state
    and hook to different periods of game execution.

    Args:
        server_name (str): the server tag to connect to
        buffer_size (int): the buffer size used by the server, the default is
                           specified by tminterface.constants.DEFAULT_BUFFER_SIZE.
                           Using a custom size requires launching TMInterface with the
                           /serversize command line parameter: TMInterface.exe /serversize=size.

    Attributes:
        server_name (str): the server tag that's used
        running (bool): whether the client is running or not
        registered (bool): whether the client is registered
        mfile (mmap.mmap): the internal mapped file used for communication
        buffer_size (int): the buffer size used for communication
        client (Client): the registered client that's controlling the server
    """
    def __init__(self, server_name='TMInterface0', buffer_size=DEFAULT_SERVER_SIZE):
        self.server_name = server_name
        self.running = True
        self.registered = False
        self.mfile = None
        self.buffer_size = buffer_size
        self.client = None
        self.empty_buffer = bytearray(self.buffer_size)
        self.thread = None
        self.request_close = False

    def register(self, client: Client) -> bool:
        """
        Registers a client on the server. 
        The server can only register one client at a time, if the client is already
        registered, the method will return False.

        This method will initially start a new thread and send a message to the server
        to register a new client. After a successful registration, :meth:`Client.on_registered`
        will be called with the instance of the TMInterface class.

        Args:
            client (Client): a Client instance to register

        Returns:
            True if registration was scheduled, False if client is already registered
        """
        if self.client is not None:
            return False

        if self.registered:
            return False

        self.registered = False
        self.client = client

        if self.thread is None:
            self.thread = threading.Thread(target=self._main_thread)
            self.thread.daemon = True
            self.thread.start()

        return True

    def close(self):
        """
        Closes the connection to the server by deregistering the current client
        and shutting down the thread for communication.

        This method will send a message to the server
        to deregister the current client.

        After a successful deregistration, :meth:`Client.on_deregistered`
        will be called with the instance of the TMInterface class.
        """
        if self.registered:
            msg = Message(MessageType.C_DEREGISTER)
            msg.write_int32(0)
            self._send_message(msg)
            self.client.on_deregistered(self)
            self.thread = None

        self.running = False

    def set_timeout(self, timeout_ms: int):
        """
        Sets the timeout window in which the client has to respond to server calls.

        The timeout specifies how long will the server wait for a response from the client.
        If the response does not arrive in this time frame, the server will automatically deregister the
        client itself.

        The timeout is specified in milliseconds, by default this is 2000ms (2 seconds). Set the timeout
        to -1 to have the server wait forever for the response.

        Args:
            timeout_ms (int): the timeout in milliseconds
        """
        msg = Message(MessageType.C_SET_TIMEOUT)
        msg.write_int32(timeout_ms)
        self._send_message(msg)
        self._wait_for_server_response()

    def set_speed(self, speed: float):
        """
        Sets the global game speed, internally this simply sets the console variable
        "speed" in an TMInterface instance.

        All characteristics of setting the global speed apply. It is not recommended
        to set the speed to high factors (such as >100), which could cause the game
        to skip running some subsystems such as the input subsystem.

        This variable does not affect simulation contexts in which debug mode is disabled.
        When debug mode is disabled (default), the game runs only the simulation subsystem.

        Args:
            speed (float): the speed to set, 1 is the default normal game speed,
                        factors <1 will slow down the game while factors >1 will speed it up
        """
        msg = Message(MessageType.C_SET_GAME_SPEED)
        msg.write_double(speed)
        self._send_message(msg)
        self._wait_for_server_response()

    def set_input_state(self, sim_clear_buffer: bool = True, **kwargs):
        """
        Sets the game input state of the vehicle. 

        Sets individual input states for the car. If successfully applied, 
        key states are guaranteed to be applied at next physics tick.
        If you want to apply an input state that happens at 500ms, call
        send this message at 490ms (one step before). 

        Note that it is not guaranteed that the game will actually process the input
        in the RUN mode. This can happen when setting the game speed to high factors
        (such as >100). This does not affect the simulation context.

        In a simulation context, the server will add new input events to the existing
        event buffer such that that the next tick has the desired input state. By default,
        all other input events are cleared. If you want to preserve existing input state & events,
        pass sim_clear_buffer=False.

        Arguments left, right, accelerate and brake are binary events. 
        To disable an action pass False and to enable it, pass True.

        Arguments steer and gas are analog events. Pass a value in the range of [-65536, 65536] to modify
        the state of these actions. You can also use the extended steer range of [-6553600, 6553600],
        note however that this range is not possible to achieve on physical hardware. This call
        is not affected by the extended_steer console variable.

        Args:
            sim_clear_buffer (bool): whether to clear the event buffer when setting
                                    input state in simulation
            **kwargs: the keyword arguments

        Keyword Args:
            left (bool): the left binary input, False = disabled, True = enabled
            right (bool): the right binary input, False = disabled, True = enabled
            accelerate (bool): the up binary input, False = disabled, True = enabled
            brake (bool): the down binary input, False = disabled, True = enabled
            steer (int): the steer analog input, in range of [-65536, 65536]
            gas (int): the gas analog input, in range of [-65536, 65536]
        """
        if self.get_context_mode() == MODE_SIMULATION and sim_clear_buffer:
            self.clear_event_buffer()

        msg = Message(MessageType.C_SET_INPUT_STATES)
        if 'left' in kwargs:
            msg.write_int32(int(kwargs['left']))
        else:
            msg.write_int32(-1)

        if 'right' in kwargs:
            msg.write_int32(int(kwargs['right']))
        else:
            msg.write_int32(-1)

        if 'accelerate' in kwargs:
            msg.write_int32(int(kwargs['accelerate']))
        else:
            msg.write_int32(-1)

        if 'brake' in kwargs:
            msg.write_int32(int(kwargs['brake']))
        else:
            msg.write_int32(-1)

        if 'steer' in kwargs:
            msg.write_int32(kwargs['steer'])
        else:
            msg.write_int32(MAXINT32)

        if 'gas' in kwargs:
            msg.write_int32(kwargs['gas'])
        else:
            msg.write_int32(MAXINT32)

        self._send_message(msg)
        self._wait_for_server_response()

    def respawn(self, sim_clear_events: bool = True):
        """
        Queues a deterministic respawn at the next race tick. This function
        will not immediately call the game to respawn the car, as TMInterface
        has to call the specific function at a specific place in the game loop.

        In a simulation context, the server will add a new input event to the existing
        event buffer such that that the car respawns at the next tick. By default,
        all other input events are cleared. If you want to preserve existing input events,
        pass sim_clear_events=False.

        The function will respawn the car to the nearest respawnable checkpoint or
        if there was no passed checkpoints, restart the race. The behaviour of this function
        also depends on the start_respawn console variable set within TMInterface.
        If start_respawn is set to true, respawning without any passed checkpoints will
        not restart the race, but only respawn the car on the start block, simulating
        online respawn behaviour.

        Args:
            sim_clear_events (bool): whether to clear all other events in simulation mode
        """
        if self.get_context_mode() == MODE_SIMULATION and sim_clear_events:
            self.clear_event_buffer()

        msg = Message(MessageType.C_RESPAWN)
        msg.write_int32(0)
        self._send_message(msg)
        self._wait_for_server_response()

    def give_up(self):
        """
        Restarts the current race.

        This function does not do anything in a simulation context.
        To rewind to the start of the race in the simulation context, use simulation states.
        """
        msg = Message(MessageType.C_GIVE_UP)
        msg.write_int32(0)
        self._send_message(msg)
        self._wait_for_server_response()

    def horn(self, sim_clear_events: bool = True):
        """
        Queues a deterministic horn at next race tick. This function
        will not immediately call the game to horn, as TMInterface
        has to call the specific function at a specific place in the game loop.

        In a simulation context, the server will add a new input event to the existing
        event buffer such that that the car horns at the next tick. By default,
        all other input events are cleared. If you want to preserve existing input events,
        pass sim_clear_events=False.

        Args:
            sim_clear_events (bool): whether to clear all other events in simulation mode
        """
        if self.get_context_mode() == MODE_SIMULATION and sim_clear_events:
            self.clear_event_buffer()

        msg = Message(MessageType.C_HORN)
        msg.write_int32(0)
        self._send_message(msg)
        self._wait_for_server_response()

    def execute_command(self, command: str):
        """
        Adds an interface command to the internal command queue.

        The command will not be immediately executed, rather, it may be executed when 
        the current queue is processed on the next game frame.

        Args:
            command (str): the command to execute
        """
        msg = Message(MessageType.C_EXECUTE_COMMAND)
        msg.write_int32(0)
        self._write_vector(msg, [ord(c) for c in command], 1)
        self._send_message(msg)
        self._wait_for_server_response()

    def remove_state_validation(self):
        """
        Makes the game validate the replay without checking if the inputs match
        the states saved in the replay, as if it was validating a replay exported
        for validation.

        Calling this method in the on_simulation_begin call will remove state
        validation from currently validated replay. After calling, TrackMania will not
        check if the simulation matches with saved states in the replay,
        therefore allowing for input modification without stopping
        the simulation prematurely.
        """
        msg = Message(MessageType.C_REMOVE_STATE_VALIDATION)
        msg.write_int32(0)
        self._send_message(msg)
        self._wait_for_server_response()

    def prevent_simulation_finish(self):
        """
        Prevents the game from stopping the simulation after a finished race.

        Calling this method in the on_checkpoint_count_changed will invalidate
        checkpoint state so that the game does not stop simulating the race.
        Internally this is simply setting the last checkpoint time to -1
        and can be also done manually in the client if additional handling
        is required.
        """
        msg = Message(MessageType.C_PREVENT_SIMULATION_FINISH)
        msg.write_int32(0)
        self._send_message(msg)
        self._wait_for_server_response()

    def rewind_to_state(self, state: SimStateData):
        """
        Rewinds to the provided simulation state.

        The method of restoring the simulation state slightly varies depending
        on the context_mode field of the SimStateData class. Some buffers
        may not be restored at all but are replaced with native game function calls.

        The simulation state is obtainable through the get_simulation_state method.
        This state can also be used to write a save state compatible file for TMInterface.
        Note that modifying important parts of the state invalidates the current race.

        To provide state restoration across game instances, TMInterface uses
        memory masks to omit restoring instance specific fields such as pointers
        or arrays.

        The method can be called in on_run_step or on_simulation_step calls.
        Note that rewinding to a state in any of these hooks will immediately
        simulate the next step after the hook. For example, rewinding to a state
        saved at race time 0, will result in the next call to on_run_step/on_simulation_step
        being at time 10. If you want to apply any immediate input state,
        make sure to apply it in the same physics step as the call to rewind_to_state.

        Args:
            state (SimStateData): the state to restore, obtained through get_simulation_state
        """
        msg = Message(MessageType.C_SIM_REWIND_TO_STATE)
        msg.write_int32(state.version)
        msg.write_int32(state.context_mode)
        msg.write_int32(state.flags)
        msg.write_buffer(state.timers)
        msg.write_buffer(state.dyna)
        msg.write_buffer(state.scene_mobil)
        msg.write_buffer(state.simulation_wheels)
        msg.write_buffer(state.plug_solid)
        msg.write_buffer(state.cmd_buffer_core)
        msg.write_buffer(state.player_info)
        msg.write_buffer(state.internal_input_state)

        msg.write_event(state.input_running_event)
        msg.write_event(state.input_finish_event)
        msg.write_event(state.input_accelerate_event)
        msg.write_event(state.input_brake_event)
        msg.write_event(state.input_left_event)
        msg.write_event(state.input_right_event)
        msg.write_event(state.input_steer_event)
        msg.write_event(state.input_gas_event)
        msg.write_uint32(state.num_respawns)

        self.__write_checkpoint_state(msg, state.cp_data)
        self._send_message(msg)
        self._wait_for_server_response()

    def set_checkpoint_state(self, data: CheckpointData):
        """
        Sets the checkpoint state of the game.
        See get_checkpoint_state to learn more about how the game stores checkpoint information.

        Args:
            data (CheckpointData): the checkpoint data
        """
        msg = Message(MessageType.C_SET_CHECKPOINT_STATE)
        self.__write_checkpoint_state(msg, data)
        self._send_message(msg)
        self._wait_for_server_response()

    def set_event_buffer(self, data: EventBufferData):
        """
        Replaces the internal event buffer used for simulation with a new one.

        If you do not modify existing inputs or do not generate all events
        beforehand, use TMInterface.set_input_state for dynamic input injection.
        See EventBufferData for more information.

        The events_duration and control_names fields are ignored in this call.

        Args:
            data (EventBufferData): the new event buffer
        """
        msg = Message(MessageType.C_SIM_SET_EVENT_BUFFER)
        for _ in range(10):
            msg.write_int32(-1)

        msg.write_int32(data.events_duration)
        events_tup = [(event.time, event.data) for event in data.events]

        self._write_vector(msg, events_tup, [4, 4])
        self._send_message(msg)
        self._wait_for_server_response()

    def get_context_mode(self) -> int:
        """
        Gets the context mode the TMInterface instance is currently in.

        The context mode is determining if the current race is in
        "run" mode, that is a normal race or "simulation" mode, which is when
        a player validates a replay.

        Returns:
            int: MODE_SIMULATION (0) if the player is in the simulation mode, MODE_RUN (1) if in a normal race
        """
        msg = Message(MessageType.C_GET_CONTEXT_MODE)
        self._send_message(msg)
        self._wait_for_server_response(False)

        self.mfile.seek(8)
        mode = self._read_int32()
        self._clear_buffer()
        return mode

    def get_checkpoint_state(self) -> CheckpointData:
        """
        Gets the current checkpoint state of the race.

        See CheckpointData for more information.

        Returns:
            CheckpointData: the object that holds the two arrays representing checkpoint state
        """
        msg = Message(MessageType.C_GET_CHECKPOINT_STATE)
        self._send_message(msg)
        self._wait_for_server_response(False)

        self.mfile.seek(4)
        error_code = self._read_int32()
        if error_code == NO_PLAYER_INFO:
            raise ServerException('Failed to get checkpoint state: no player info available')

        data = self._read_checkpoint_state()

        self._clear_buffer()
        return data

    def get_simulation_state(self) -> SimStateData:
        """
        Gets the current simulation state of the race.

        The method can be called in on_run_step or on_simulation_step calls.
        See SimStateData for more information.

        Returns:
            SimStateData: the object holding the simulation state
        """
        msg = Message(MessageType.C_SIM_GET_STATE)
        self._send_message(msg)
        self._wait_for_server_response(False)

        self.mfile.seek(4)
        error_code = self._read_int32()

        state = SimStateData()
        state.version = self._read_int32()
        state.context_mode = self._read_int32()
        state.flags = self._read_uint32()
        state.timers = bytearray(self.mfile.read(TIMERS_SIZE))
        state.dyna = bytearray(self.mfile.read(DYNA_SIZE))
        state.scene_mobil = bytearray(self.mfile.read(SCENE_MOBIL_SIZE))
        state.simulation_wheels = bytearray(self.mfile.read(SIMULATION_WHEELS_SIZE))
        state.plug_solid = bytearray(self.mfile.read(PLUG_SOLID_SIZE))
        state.cmd_buffer_core = bytearray(self.mfile.read(CMD_BUFFER_CORE_SIZE))
        state.player_info = bytearray(self.mfile.read(PLAYER_INFO_SIZE))
        state.internal_input_state = bytearray(self.mfile.read(INPUT_STATE_SIZE))

        state.input_running_event = self._read_event()
        state.input_finish_event = self._read_event()
        state.input_accelerate_event = self._read_event()
        state.input_brake_event = self._read_event()
        state.input_left_event = self._read_event()
        state.input_right_event = self._read_event()
        state.input_steer_event = self._read_event()
        state.input_gas_event = self._read_event()
        state.num_respawns = self._read_uint32()

        if error_code == NO_PLAYER_INFO:
            raise ServerException('Failed to get checkpoint state: no player info available')

        state.cp_data = self._read_checkpoint_state()

        self._clear_buffer()
        return state

    def get_event_buffer(self) -> EventBufferData:
        """
        Gets the internal event buffer used to hold player inputs in run or simulation mode.
        If the server is in the run mode (that is, in a normal race controlled by the player),
        this method returns the inputs of the current race. Note that new inputs will be added
        to the buffer as the player or TMInterface injects inputs into the game.

        See EventBufferData for more information.

        Returns:
            EventBufferData: the event buffer holding all the inputs of the current simulation
        """
        msg = Message(MessageType.C_SIM_GET_EVENT_BUFFER)
        self._send_message(msg)
        self._wait_for_server_response(False)

        self.mfile.seek(4)
        error_code = self._read_uint32()
        if error_code == NO_EVENT_BUFFER:
            raise ServerException('Failed to get event buffer: no event buffer available')

        names = [None] * 10
        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_RACE_START_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_RACE_FINISH_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_ACCELERATE_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_BRAKE_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_LEFT_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_RIGHT_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = ANALOG_STEER_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = ANALOG_ACCELERATE_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_RESPAWN_NAME

        _id = self._read_int32()
        if _id != -1:
            names[_id] = BINARY_HORN_NAME

        data = EventBufferData(self._read_uint32())
        data.control_names = names
        event_data = self.__read_vector([4, 4])
        for item in event_data:
            ev = Event(item[0], item[1])
            data.events.append(ev)

        self._clear_buffer()
        return data

    def clear_event_buffer(self):
        """
        Clears the current event buffer used for simulation, leaving
        the race running event in the buffer.

        A race running event should always be present in the buffer, to
        make the game start the race.
        """
        event_buffer = self.get_event_buffer()
        event_buffer.clear()
        self.set_event_buffer(event_buffer)

    def set_simulation_time_limit(self, time: int):
        """
        Sets the time limit of the simulation.

        This allows for setting an arbitrary time limit for the running
        simulation, making the game stop the simulation after the provided
        time limit is exhausted.

        By default, this limit is set to the finish time of the original replay
        (taken from events duration found in the events buffer).

        Note that setting the limit to a large value will extend the simulation
        to that limit, even after the race is finished. Make sure to manage
        finishing the race according to your application (e.g by rewinding to a state).

        To reset the time to the original limit, pass -1.
        This call applies only to the simulation context.

        Args:
            time (int): the time at which the game stops simulating, pass -1 to reset
                        to the original value
        """
        msg = Message(MessageType.C_SIM_SET_TIME_LIMIT)
        msg.write_int32(time)
        self._send_message(msg)
        self._wait_for_server_response()

    def register_custom_command(self, command: str):
        """
        Registers a custom command within the console.

        This function allows you to implement a custom command that is registered within TMInterface's console.
        When executing the command, the :meth:`Client.on_custom_command` method will be called with additional
        arguments such as the time range and processed arguments list.

        It is completely up to the command implementation to process the time range and additional
        arguments supplied in the on_custom_command hook. Quoted arguments such as filenames will
        be automatically joined into one argument, even if they contain spaces. 

        A console command is not immediately executed after submitting it to the console.
        TMInterface executes commands asynchronously, processing a fixed amount of commands
        each frame. This is done to prevent the game from hanging when loading scripts with
        1000's of commands.

        Use the log() method to output any info about the execution of your command.

        Args:
            command (str): the command to register, the command cannot contain spaces
        """
        msg = Message(MessageType.C_REGISTER_CUSTOM_COMMAND)
        msg.write_int32(0)
        self._write_vector(msg, [ord(c) for c in command], 1)
        self._send_message(msg)
        self._wait_for_server_response(False)

        self.mfile.seek(4)
        error_code = self._read_int32()
        if error_code == COMMAND_ALREADY_REGISTERED:
            raise ServerException(f'Failed to register custom command: {command} is already registered')

        self._clear_buffer()

    def log(self, message: str, severity='log'):
        """
        Prints a message in TMInterface's console.

        You can specify the severity of the command to highlight the line in a different color.

        Args:
            message (str): the message to print
            severity (str): one of: "log", "success", "warning", "error", the message severity
        """
        severity_id = 0
        if severity == 'success':
            severity_id = 1
        elif severity == 'warning':
            severity_id = 2
        elif severity == 'error':
            severity_id = 3

        msg = Message(MessageType.C_LOG)
        msg.write_int32(severity_id)
        self._write_vector(msg, [ord(c) for c in message], 1)
        self._send_message(msg)
        self._wait_for_server_response()

    def _on_bruteforce_validate_call(self, msgtype: MessageType):
        info = BFEvaluationInfo()
        info.phase = BFPhase(self._read_int32())
        info.target = BFTarget(self._read_int32())
        info.time = self._read_int32()
        info.modified_inputs_num = self._read_int32()
        info.inputs_min_time = self._read_int32()
        info.inputs_max_time = self._read_int32()
        info.max_steer_diff = self._read_int32()
        info.max_time_diff = self._read_int32()
        info.override_stop_time = self._read_int32()
        info.search_forever = bool(self._read_int32())
        info.inputs_extend_steer = bool(self._read_int32())

        resp = self.client.on_bruteforce_evaluate(self, info)
        if not resp:
            resp = BFEvaluationResponse()

        msg = Message(MessageType.C_PROCESSED_CALL)
        msg.write_int32(msgtype)
        msg.write_int32(resp.decision)
        msg.write_int32(resp.rewind_time)
        self._send_message(msg)

    def __write_checkpoint_state(self, msg: Message, data: CheckpointData):
        msg.write_int32(0)  # reserved
        if self._write_vector(msg, data.cp_states, 4):
            self._write_vector(msg, data.cp_times, [4, 4])

    def _read_checkpoint_state(self):
        self._read_int32()  # reserved
        cp_states = self.__read_vector(4)
        cp_times = self.__read_vector([4, 4])

        data = CheckpointData(cp_states, cp_times)
        return data

    def _read_event(self):
        time = self._read_uint32()
        data = self._read_int32()
        return Event(time, data)

    def _write_vector(self, msg: Message, vector: list, field_sizes):
        is_list = isinstance(field_sizes, list)
        if is_list:
            item_size = sum(field_sizes)
        else:
            item_size = field_sizes

        if len(msg) + 4 > self.buffer_size:
            return False

        vsize = len(vector)
        msgsize = len(msg) + 4 + item_size * vsize
        if msgsize > self.buffer_size:
            msg.write_int32(0)
            msg.error_code = RESPONSE_TOO_LONG
            return True

        msg.write_int32(vsize)
        if is_list:
            for elem in vector:
                for i, field in enumerate(elem):
                    msg.write_int(field, field_sizes[i])
        else:
            for elem in vector:
                msg.write_int(elem, field_sizes)

        return True

    def __read_vector(self, field_sizes: Tuple[int, list]) -> list:
        if self.mfile.tell() + 4 > self.buffer_size:
            return []

        size = self._read_int32()
        vec = []
        for _ in range(size):
            if isinstance(field_sizes, list):
                tup = []
                for size in field_sizes:
                    tup.append(self._read_int(size))

                vec.append(tuple(tup))
            else:
                vec.append(self._read_int(field_sizes))

        return vec

    def _main_thread(self):
        while self.running:
            if not self._ensure_connected():
                time.sleep(0)
                continue

            if not self.registered:
                msg = Message(MessageType.C_REGISTER)
                self._send_message(msg)
                self._wait_for_server_response()
                self.registered = True

            self._process_server_message()
            time.sleep(0)

    def _process_server_message(self):
        if self.mfile is None:
            return

        self.mfile.seek(0)
        msgtype = self._read_int32()
        if msgtype & 0xFF00 == 0:
            return

        msgtype &= 0xFF

        # error_code = self.__read_int32()
        self._skip(4)

        if msgtype == MessageType.S_SHUTDOWN:
            self.close()
            self.client.on_shutdown(self)
        elif msgtype == MessageType.S_ON_RUN_STEP:
            _time = self._read_int32()
            self.client.on_run_step(self, _time)
            self._respond_to_call(msgtype)
        elif msgtype == MessageType.S_ON_SIM_BEGIN:
            self.client.on_simulation_begin(self)
            self._respond_to_call(msgtype)
        elif msgtype == MessageType.S_ON_SIM_STEP:
            _time = self._read_int32()
            self.client.on_simulation_step(self, _time)
            self._respond_to_call(msgtype)
        elif msgtype == MessageType.S_ON_SIM_END:
            result = self._read_int32()
            self.client.on_simulation_end(self, result)
            self._respond_to_call(msgtype)
        elif msgtype == MessageType.S_ON_CHECKPOINT_COUNT_CHANGED:
            current = self._read_int32()
            target = self._read_int32()
            self.client.on_checkpoint_count_changed(self, current, target)
            self._respond_to_call(msgtype)
        elif msgtype == MessageType.S_ON_LAPS_COUNT_CHANGED:
            current = self._read_int32()
            self.client.on_laps_count_changed(self, current)
            self._respond_to_call(msgtype)
        elif msgtype == MessageType.S_ON_BRUTEFORCE_EVALUATE:
            self._on_bruteforce_validate_call(msgtype)
        elif msgtype == MessageType.S_ON_REGISTERED:
            self.registered = True
            self.client.on_registered(self)
            self._respond_to_call(msgtype)
        elif msgtype == MessageType.S_ON_CUSTOM_COMMAND:
            _from = self._read_int32()
            to = self._read_int32()
            n_args = self._read_int32()
            command = self._read_string()
            args = []
            for _ in range(n_args):
                args.append(self._read_string())

            self.client.on_custom_command(self, _from, to, command, args)
            self._respond_to_call(msgtype)

    def _ensure_connected(self):
        if self.mfile is not None:
            return True

        try:
            self.mfile = mmap.mmap(-1, self.buffer_size, tagname=self.server_name)
            return True
        except Exception as e:
            self.client.on_client_exception(self, e)

        return False

    def _wait_for_server_response(self, clear: bool = True):
        if self.mfile is None:
            return

        self.mfile.seek(0)
        while self._read_int32() != MessageType.S_RESPONSE | 0xFF00:
            self.mfile.seek(0)
            time.sleep(0)

        if clear:
            self._clear_buffer()

    def _respond_to_call(self, msgtype: int):
        msg = Message(MessageType.C_PROCESSED_CALL)
        msg.write_int32(msgtype)
        self._send_message(msg)

    def _send_message(self, message: Message):
        if self.mfile is None:
            return

        data = message.to_data()
        self.mfile.seek(0)
        self.mfile.write(data)

        self.mfile.seek(1)
        self.mfile.write(bytearray([0xFF]))

    def _clear_buffer(self):
        self.mfile.seek(0)
        self.mfile.write(self.empty_buffer)

    def _read(self, num_bytes: int, typestr: str):
        arr = self.mfile.read(num_bytes)
        try:
            return struct.unpack(typestr, arr)[0]
        except Exception as e:
            self.client.on_client_exception(self, e)
            return 0

    def _read_int(self, size):
        if size == 1:
            return self._read_uint8()
        if size == 2:
            return self._read_uint16()
        elif size == 4:
            return self._read_int32()

        return 0

    def _read_uint8(self):
        return self._read(1, 'B')

    def _read_int32(self):
        return self._read(4, 'i')

    def _read_uint32(self):
        return self._read(4, 'I')

    def _read_uint16(self):
        return self._read(2, 'H')

    def _read_string(self):
        chars = [chr(b) for b in self.__read_vector(1)]
        return ''.join(chars)

    def _skip(self, n):
        self.mfile.seek(self.mfile.tell() + n)



from distutils.core import setup
setup(
    name = 'tminterface',
    packages = ['tminterface'],
    version = '0.4',
    license='GPL3',
    description = 'A client for TMInterface, a TrackMania TAS tool',
    author = 'Adam Bieńkowski',
    author_email = 'donadigos159@gmail.com',
    url = 'https://github.com/donadigo/TMInterfaceClientPython',
    download_url = 'https://github.com/donadigo/TMInterfaceClientPython/archive/refs/tags/0.1.tar.gz',
    keywords = ['TMInterface', 'client', 'TrackMania'],
    install_requires=[
        'numpy',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)


from dataclasses import dataclass
from enum import IntEnum
from io import IOBase

BOT_COMMANDS = ['press', 'rel', 'steer', 'gas']
BOT_INPUT_TYPES = ['up', 'down', 'left', 'right', 'enter', 'delete', 'horn', 'steer', 'gas']


class InputType(IntEnum):
    """
    The InputType enum represents an input type that is coupled with its state.
    """
    UP = 0
    DOWN = 1
    LEFT = 2
    RIGHT = 3
    RESPAWN = 4
    RESET = 5
    HORN = 6
    STEER = 7
    GAS = 8
    UNKNOWN = 9

    @staticmethod
    def from_str(s: str):
        """
        Converts a script action string into an InputType.

        If the action is invalid, InputType.UNKNOWN will be returned.

        Args:
            s (str): the string to convert

        Returns:
            InputType: the converted input type
        """
        s = s.lower()
        if s not in BOT_INPUT_TYPES:
            return InputType.UNKNOWN

        return InputType(BOT_INPUT_TYPES.index(s))

    def to_str(self) -> str:
        if int(self) < len(BOT_INPUT_TYPES):
            return BOT_INPUT_TYPES[int(self)]
        else:
            return 'unknown'


class BaseCommand:
    """
    The BaseCommand class is a base class for all command classes such as Command, TimedCommand and InputCommand.
    """
    def to_script(self) -> str:
        """
        Converts a command to a script string line.

        Returns:
            str: the script snippet that represents this command
        """
        pass


@dataclass
class Command(BaseCommand):
    """
    A Command represents an immediate command that is executed immediately by TMInterface whenever it's encountered.
    """
    args: list

    def to_script(self) -> str:
        return ' '.join(self.args)


@dataclass
class InputCommand(BaseCommand):
    """
    The InputCommand class specifically represents a command that is used to inject any kind of input into the game.

    An input does not contain any arguments. Instead, the class defines an input with its type and state.
    InputCommand's can be converted from an instance of TimedCommand.

    InputCommand's do not need to be stored to describe a TMInterface script, they are however automatically added
    by the CommandList for an easy access & manipulation of the input sequence.
    """
    timestamp: int
    input_type: InputType
    state: int

    def to_script(self) -> str:
        if self.input_type == InputType.STEER or self.input_type == InputType.GAS:
            return f'{self.timestamp} {self.input_type.to_str()} {self.state}'
        elif self.input_type == InputType.UNKNOWN:
            return f'# {self.timestamp} [unknown] {int(self.state)}'
        else:
            action = 'press' if self.state else 'rel'
            return f'{self.timestamp} {action} {self.input_type.to_str()}'


@dataclass
class TimedCommand(Command):
    """
    A TimedCommand describes any command that is executed at a specific timestamp.

    The TimedCommand can represent any command, including any input commands.
    A command with a ranged timestamp will be always converted to two TimedCommand instances,
    where the earliest command will have is_ending set to False and the latest, to True.
    """
    timestamp: int
    is_ending: bool

    def to_input_command(self) -> InputCommand:
        """
        Converts a TimedCommand to an InputCommand if possible.

        If the conversion fails or the TimedCommand is not a valid input command, None is returned.

        Returns:
            InputCommand: the converted InputCommand, or None if the conversion failed
        """
        if len(self.args) < 2 or self.args[0].lower() not in BOT_COMMANDS:
            return None

        state = 0
        action = self.args[0].lower()
        input_type = InputType.UNKNOWN
        if action == 'press' or action == 'rel':
            if action == 'press' and not self.is_ending:
                state = 1

            input_type = InputType.from_str(self.args[1])
        elif action == 'steer' or action == 'gas':
            input_type = InputType.from_str(action)

            if not self.is_ending:
                try:
                    state = int(self.args[1])
                except ValueError:
                    return None
        else:
            return None

        return InputCommand(self.timestamp, input_type, state)

    def to_script(self) -> str:
        input_command = self.to_input_command()
        if input_command:
            return input_command.to_script()

        return f'{self.timestamp} {super().to_script()}'


class CommandList(object):
    """
    A CommandList represents a list of TMInterface commands usually forming a script which can contain immediate
    and timed commands.

    A CommandList can be loaded by providing a file handle to an existing script file or from a string.
    You can also construct an empty CommandList to add your own commands to then convert them 
    into a valid TMInterface script.

    If a resource is provided, the class will attempt to parse all of its contents into immediate and timed commands.
    You can use CommandList.to_script() to convert all the commands back into a valid TMInterface script.
    If any command cannot be converted, it will be commented out.

    The class fully supports parsing commands with quoted arguments and inline comments and can be used
    to genereate new script files.

    Args:
        obj: the resource that needs to be parsed, either:

             a file handle opened with open()

             a string containing the command list

             None to create an empty list

    Attributes:
        commands (list): the list containing all immediate commands
        timed_commands (list): the list containing all timed commands, including input commands
        content (str): the script string that was used to construct the CommandList
    """
    def __init__(self, obj=None):
        self.commands = []
        self.timed_commands = []
        self.content = None

        if obj:
            if isinstance(obj, IOBase):
                self.content = obj.read()
            else:
                self.content = obj

            self._parse()

    def _parse(self):
        for line in self.content.split('\n'):
            line = line.split('#')[0].strip()
            if not line or line.startswith('#'):
                continue

            for command in CommandList._split_input(line):
                self._parse_command(command)

    def _parse_command(self, command):
        args = CommandList._split_command_args(command)
        if not args:
            return

        _from, _to = CommandList.parse_time_range(args[0])
        if _from != -1:
            self.add_command(TimedCommand(args[1:], _from, False))

            if _to != -1:
                self.add_command(TimedCommand(args[1:], _to, True))
        else:
            self.commands.append(Command(args))

    def sorted_timed_commands(self) -> list:
        """
        Returns all timed commands sorted in ascending order (stable).

        Returns:
            list: timed commands sorted in ascending order
        """
        return sorted(self.timed_commands, key=lambda command: command.timestamp)

    def add_command(self, command: BaseCommand):
        """
        Adds a command to the CommandList, converting it to an InputCommand if possible.

        The command will be added to the commands list if it is of type Command.
        If the command is a TimedCommand, it will first be attempted to convert it
        to an InputCommand. If the conversion fails, it is added without any conversions.
        If the command is an InputCommnad, it is added to the timed_commands list.

        Args:
            command (BaseCommand): the command to be added
        """
        if type(command) == Command:
            self.commands.append(command)
        elif type(command) == TimedCommand:
            input_command = command.to_input_command()
            if input_command:
                self.timed_commands.append(input_command)
            else:
                self.timed_commands.append(command)
        elif type(command) == InputCommand:
            self.timed_commands.append(command)

    def to_script(self) -> str:
        """
        Converts all immediate and timed commands to a valid TMInterface script.

        Returns:
            str: the string representing the TMInterface script, one command per line
        """
        script = ''
        for command in self.commands:
            script += f'{command.to_script()}\n'

        for command in self.sorted_timed_commands():
            script += f'{command.to_script()}\n'

        return script

    def clear(self):
        """
        Clears all commands from the command list.
        """
        self.commands.clear()
        self.timed_commands.clear()

    @staticmethod
    def _split_input(command_input: str) -> list:
        in_quotes = False
        commands = []
        offset = 0
        for i, c in enumerate(command_input):
            if c == '\"':
                in_quotes = not in_quotes

            if not in_quotes and c == ';':
                commands.append(command_input[offset:i])
                offset = i + 1

        if len(command_input) - offset > 0:
            commands.append(command_input[offset:])

        return commands

    @staticmethod
    def _split_command_args(command: str) -> list:
        args = []
        offset = 0
        i = 0
        while i < len(command):
            if command[i] == ' ':
                if command[offset] != ' ':
                    args.append(command[offset:i])

                offset = i + 1
            elif command[i] == '\"':
                i += 1
                closing = command.find('\"', i)
                if closing != -1:
                    if closing - i > 0:
                        args.append(command[i:closing])

                    i = closing
                    offset = i + 1

            i += 1

        if len(command) - offset > 0:
            args.append(command[offset:])

        return args

    @staticmethod
    def parse_time_range(range_str: str) -> tuple:
        """
        Parses a time range.

        Parses a time range or a single timestamp, returning a tuple
        with two elements (from, to).

        If the parsed time range consists only of one timestamp, to is set to -1.
        If from > to, the two integers are swapped.

        Args:
            range_str (str): the time range to parse

        Returns:
            tuple: a tuple of two int's (from, to)
        """
        timestamps = range_str.split('-', 1)
        timestamps_len = len(timestamps)
        if timestamps_len == 1:
            return CommandList.parse_time(timestamps[0]), -1
        elif timestamps_len == 2:
            _from = CommandList.parse_time(timestamps[0])
            _to = CommandList.parse_time(timestamps[1])
            if _from > _to:
                _from, _to = _to, _from

            return _from, _to

        return -1, -1

    @staticmethod
    def _parse_seconds(time_str: str) -> int:
        tokens = time_str.split('.', 1)
        if len(tokens) < 2:
            return -1

        if not tokens[0] or not tokens[1]:
            return -1

        tokens[1] = tokens[1][:2].ljust(2, '0')

        try:
            seconds = int(tokens[0])
            milliseconds = int(tokens[1])
            return seconds * 1000 + milliseconds * 10
        except ValueError:
            return -1

    @staticmethod
    def parse_time(time_str: str) -> int:
        """
        Parses a singular timestamp which is either a number or a formatted time.

        Parses a string like "947120" or "15:47.12" to an integer time in milliseconds.

        Args:
            time_str (str): the time string to parse

        Returns:
            int: the time representing the string, -1 if parsing fails
        """
        if '.' not in time_str:
            try:
                return int(time_str)
            except ValueError:
                return -1

        tokens = time_str.split(':', 2)
        if not tokens:
            return -1

        tokens_len = len(tokens)
        if tokens_len == 1:
            return CommandList._parse_seconds(time_str)
        elif tokens_len == 2:
            try:
                minutes = int(tokens[0])
            except ValueError:
                return -1

            seconds = CommandList._parse_seconds(tokens[1])
            if seconds == -1:
                return -1

            return minutes * 60000 + seconds
        elif tokens_len == 3:
            try:
                hours = int(tokens[0])
                minutes = int(tokens[1])
            except ValueError:
                return - 1

            seconds = CommandList._parse_seconds(tokens[2])
            if seconds == -1:
                return -1

            return hours * 3600000 + minutes * 60000 + seconds

        return -1
